@@ -1,4 +1,5 @@
 
+#include <stddef.h>
 #include <stdint.h>
 #include <stdlib.h>
 #include <Rdefines.h>
@@ -80,22 +81,27 @@ SEXP BtoTPCmsm(
 	const char *name3 = CHAR( STRING_ELT(statenames, 2) );
 	i = strlen(name1), j = strlen(name2), k = strlen(name3);
 	char *name11 = (char*)malloc( (i+i+2)*sizeof(char) );
+	if (name11 == NULL) error("BtoTPCmsm: No more memory\n");
 	strcpy(name11, name1);
 	strcpy(&name11[i], " ");
 	strcpy(&name11[i+1], name1);
 	char *name12 = (char*)malloc( (i+j+2)*sizeof(char) );
+	if (name12 == NULL) error("BtoTPCmsm: No more memory\n");
 	strcpy(name12, name1);
 	strcpy(&name12[i], " ");
 	strcpy(&name12[i+1], name2);
 	char *name13 = (char*)malloc( (i+k+2)*sizeof(char) );
+	if (name13 == NULL) error("BtoTPCmsm: No more memory\n");
 	strcpy(name13, name1);
 	strcpy(&name13[i], " ");
 	strcpy(&name13[i+1], name3);
 	char *name22 = (char*)malloc( (j+j+2)*sizeof(char) );
+	if (name22 == NULL) error("BtoTPCmsm: No more memory\n");
 	strcpy(name22, name2);
 	strcpy(&name22[j], " ");
 	strcpy(&name22[j+1], name2);
 	char *name23 = (char*)malloc( (j+k+2)*sizeof(char) );
+	if (name23 == NULL) error("BtoTPCmsm: No more memory\n");
 	strcpy(name23, name2);
 	strcpy(&name23[j], " ");
 	strcpy(&name23[j+1], name3);
