@@ -6,13 +6,14 @@
 #include <omp.h>
 #endif
 #include <stdlib.h>
-#include <time.h>
 #include <Rdefines.h>
 #include <Rmath.h>
 #include "defines.h"
-#include "boot.h"
 #include "get.h"
 #include "kmsurv.h"
+#include "RngStream.h"
+#include "RngArray.h"
+#include "RngBoot.h"
 #include "rthreads.h"
 #include "sort.h"
 #include "splines.h"
@@ -21,6 +22,13 @@
 #include "wtypefunc.h"
 #include "window.h"
 
+typedef struct {
+	int *sample0, *sample1, *unique0;
+	double *K, *MX;
+	double *a, *b, *c;
+	double *E0B, *E1B;
+} transLSW;
+
 #define weights \
 	order_di(T1, E1, sample0, *len, FALSE, FALSE, TRUE, a, unique0); /* get permuation */ \
 	for (aux[0] = 1, i = 0; i < *len; i++) { /* loop through the sample until last index is reached */ \
@@ -33,7 +41,7 @@
 #define mean \
 	order_d(T2, sample0, *len, FALSE, FALSE, a); /* get permuation */ \
 	for (i = 0; i < u0; i++) { \
-		kfunc(T1, SW, sample0, &T1[unique0[i]], &h1, K); /* compute weights */ \
+		kfunc(T1, &SW, sample0, &T1[unique0[i]], &h1, K); /* compute weights */ \
 		wkmsurv(len, T2, E, K, sample0, len, K); /* compute conditional survival probabilities vector */ \
 		MX[unique0[i]] = (1-K[sample0[0]])*T2[sample0[0]]; /* initialize mean */ \
 		for (j = 1; j < *len; j++) { \
@@ -56,7 +64,7 @@
 
 /*
 Author:
-	Artur Agostinho Araujo <artur.stat@gmail.com>
+	Artur Araujo <artur.stat@gmail.com>
 
 Description:
 	Computes the kernel bandwidth used by LS estimators
@@ -70,19 +78,20 @@ Parameters:
 	T2[in]			pointer to T2 first element.
 	index[in]		pointer to index first element.
 	len[in]			pointer to length of index,
-						which must be lower or equal
-						than the length of vectors
-						T1, E1, S, E and T2.
+					which must be lower or equal
+					than the length of vectors
+					T1, E1, S, E and T2.
 	h[in]			pointer to h first element.
 	nh[in]			pointer to number of bandwidth
-						values to test by cross-validation.
+					values to test by cross-validation.
 	ncv[in]			pointer to number of
-						cross-validation samples.
+					cross-validation samples.
 	cvfull[in]		pointer to integer indicating if full
-						cross-validation is to be done.
+					cross-validation is to be done.
 	kfunc[in]		pointer to kernel density function.
-	b[in]			pointer to row index.
 	H[out]			pointer to H first element.
+	t[in]			pointer to thread number.
+	WORK[out]		pointer to array of transLSW structures.
 
 Return value:
 	This function doesn't return a value.
@@ -90,7 +99,6 @@ Return value:
 Remarks:
 	Vectors T1, E1, S, E and T2 must have the same length.
 	Cubic spline interpolation is used.
-	If (*b < 1) the code runs in parallel among the available threads.
 */
 
 static void crossValid(
@@ -106,8 +114,9 @@ static void crossValid(
 	CintCP ncv,
 	CintCP cvfull,
 	Kfunc kfunc,
-	CintCP b,
-	double H[2])
+	double H[2],
+	CintCP t,
+	transLSW *WORK)
 {
 	if (h[1] == h[0] && h[3] == h[2]) {
 		H[0] = h[0];
@@ -116,28 +125,29 @@ static void crossValid(
 	}
 	register int x, y, i, j;
 	const int method = 1;
-	unsigned int iseed;
-	int u0;
+	unsigned long iseed[6];
+	int u0, tid = *t;
 	double h0, h1, aux[3], cv0, cv1, cv2, sum;
-	stype *SW = (stype*)malloc( sizeof(stype) ); // allocate memory for stype structure
-	SW->type = INT_PTR; // type is an int pointer
-	SW->ptr.integer = E1; // hold E1 pointer in ptr union
-	SW->length = *len; // hold length of array
+	stype SW;
+	SW.type = INT_PTR; // type is an int pointer
+	SW.ptr.integer = E1; // hold E1 pointer in ptr union
+	SW.length = *len; // hold length of array
 	#ifdef _OPENMP
-	#pragma omp parallel if(*b < 1) num_threads(global_num_threads) private(x, y, i, j, iseed, u0, h1, aux, cv1, cv2, sum)
+	#pragma omp parallel if( !omp_in_parallel() ) num_threads(global_num_threads) firstprivate(tid) private(x, y, i, j, iseed, u0, h1, aux, cv1, cv2, sum)
 	#endif
 	{
-		int *sample0 = (int*)malloc( *len*sizeof(int) ); // allocate memory block
-		int *sample1 = (int*)malloc( *len*sizeof(int) ); // allocate memory block
-		double *K = (double*)malloc( *len*sizeof(double) ); // allocate memory block
-		double *MX = (double*)malloc( *len*sizeof(double) ); // allocate memory block
-		int *unique0 = (int*)malloc( *len*sizeof(int) ); // allocate memory block
-		double *a = (double*)malloc( *len*3*sizeof(double) ); // allocate memory block
 		#ifdef _OPENMP
-		iseed = (unsigned int)time(NULL) ^ (unsigned int)omp_get_thread_num(); // save per thread seed
-		#else
-		iseed = (unsigned int)time(NULL);
+		if (omp_get_num_threads() != 1) tid = omp_get_thread_num(); // use the correct thread number
 		#endif
+		int *sample0 = WORK[tid].sample0;
+		int *sample1 = WORK[tid].sample1;
+		double *K = WORK[tid].K;
+		double *MX = WORK[tid].MX;
+		int *unique0 = WORK[tid].unique0;
+		double *a = WORK[tid].a;
+		double *b = WORK[tid].b;
+		double *c = WORK[tid].c;
+		RngStream_GetState(RngArray[tid], iseed); // save per thread seed
 		if (h[1] == h[0]) {
 			#ifdef _OPENMP
 			#pragma omp single
@@ -146,7 +156,6 @@ static void crossValid(
 				H[0] = h[0];
 			}
 		} else {
-			if (*b < 1) srand(iseed); // set seed
 			#ifdef _OPENMP
 			#pragma omp single
 			#endif
@@ -160,8 +169,8 @@ static void crossValid(
 			for (x = 0; x < *nh; x++) {
 				h1 = h[0]+h0*x;
 				for (cv1 = 0, y = 0; y < *ncv; y++) {
-					boot_i(sample0, len); // simulate first random sample
-					boot_i(sample1, len); // simulate second random sample
+					boot_i(RngArray[tid], len, sample0); // simulate first random sample
+					boot_i(RngArray[tid], len, sample1); // simulate second random sample
 					for (i = 0; i < *len; i++) {
 						sample0[i] = index[sample0[i]];
 						sample1[i] = index[sample1[i]];
@@ -169,8 +178,8 @@ static void crossValid(
 					order_d(T1, sample0, *len, FALSE, FALSE, a); // get permuation
 					uniqueI(T1, sample0, len, unique0, &u0); // compute unique index
 					mean // compute mean vector
-					spline_coefI(&method, T1, MX, unique0, &u0, a, a+*len, a+*len*2); // compute spline coefficients
-					spline_evalI(&method, T1, MX, unique0, &u0, a, a+*len, a+*len*2, T1, K, sample1, len); // interpolate mean from unique0 to sample1
+					spline_coefI(&method, T1, MX, unique0, &u0, a, b, c); // compute spline coefficients
+					spline_evalI(&method, T1, MX, unique0, &u0, a, b, c, T1, K, sample1, len); // interpolate mean from unique0 to sample1
 					cverror // compute cross-validation error term
 					cv1 += cv2 / *ncv; // compute cross-validation error
 				}
@@ -194,9 +203,9 @@ static void crossValid(
 					H[1] = h[2];
 				}
 			} else {
-				if (*b < 1) srand(iseed); // set seed
-				double *E0B = (double*)malloc( *len*sizeof(double) ); // allocate memory block
-				double *E1B = (double*)malloc( *len*sizeof(double) ); // allocate memory block
+				RngStream_SetSeed(RngArray[tid], iseed); // restore seed
+				double *E0B = WORK[tid].E0B;
+				double *E1B = WORK[tid].E1B;
 				#ifdef _OPENMP
 				#pragma omp single
 				#endif
@@ -210,8 +219,8 @@ static void crossValid(
 				for (x = 0; x < *nh; x++) {
 					h1 = h[2]+h0*x;
 					for (cv1 = 0, y = 0; y < *ncv; y++) {
-						boot_i(sample0, len); // simulate first random sample
-						boot_i(sample1, len); // simulate second random sample
+						boot_i(RngArray[tid], len, sample0); // simulate first random sample
+						boot_i(RngArray[tid], len, sample1); // simulate second random sample
 						for (i = 0; i < *len; i++) {
 							sample0[i] = index[sample0[i]];
 							sample1[i] = index[sample1[i]];
@@ -221,9 +230,9 @@ static void crossValid(
 						#define h1 H[0]
 						mean // compute mean vector
 						#undef h1
-						spline_coefI(&method, T1, MX, unique0, &u0, a, a+*len, a+*len*2); // compute spline coefficients
-						spline_evalI(&method, T1, MX, unique0, &u0, a, a+*len, a+*len*2, T1, MX, sample0, len); // interpolate mean from unique0 to sample0
-						spline_evalI(&method, T1, MX, unique0, &u0, a, a+*len, a+*len*2, T1, K, sample1, len); // interpolate mean from unique0 to sample1
+						spline_coefI(&method, T1, MX, unique0, &u0, a, b, c); // compute spline coefficients
+						spline_evalI(&method, T1, MX, unique0, &u0, a, b, c, T1, MX, sample0, len); // interpolate mean from unique0 to sample0
+						spline_evalI(&method, T1, MX, unique0, &u0, a, b, c, T1, K, sample1, len); // interpolate mean from unique0 to sample1
 						for (i = 0; i < *len; i++) {
 							E0B[sample0[i]] = T2[sample0[i]]-MX[sample0[i]]; // compute error
 							E1B[sample1[i]] = T2[sample1[i]]-K[sample1[i]]; // compute error
@@ -231,8 +240,8 @@ static void crossValid(
 						#define T2 E0B
 						mean // compute mean vector
 						#undef T2
-						spline_coefI(&method, T1, MX, unique0, &u0, a, a+*len, a+*len*2); // compute spline coefficients
-						spline_evalI(&method, T1, MX, unique0, &u0, a, a+*len, a+*len*2, T1, K, sample1, len); // interpolate mean from unique0 to sample1
+						spline_coefI(&method, T1, MX, unique0, &u0, a, b, c); // compute spline coefficients
+						spline_evalI(&method, T1, MX, unique0, &u0, a, b, c, T1, K, sample1, len); // interpolate mean from unique0 to sample1
 						for (i = 0; i < *len; i++) MX[sample1[i]] = T1[sample1[i]]+E1B[sample1[i]]; // compute total time
 						#define S MX
 						#define T2 E1B
@@ -251,8 +260,6 @@ static void crossValid(
 						}
 					}
 				}
-				free(E0B); // free memory block
-				free(E1B); // free memory block
 			}
 		} else {
 			#ifdef _OPENMP
@@ -262,20 +269,13 @@ static void crossValid(
 				H[1] = H[0];
 			}
 		}
-		free(sample0); // free memory block
-		free(sample1); // free memory block
-		free(K); // free memory block
-		free(MX); // free memory block
-		free(unique0); // free memory block
-		free(a); // free memory block
 	}
-	free(SW); // free memory block
 	return;
 } // crossValid
 
 /*
 Author:
-	Artur Agostinho Araujo <artur.stat@gmail.com>
+	Artur Araujo <artur.stat@gmail.com>
 
 Description:
 	Computes the mean and variance vectors respectively
@@ -288,14 +288,15 @@ Parameters:
 	E[in]			pointer to E first element.
 	index[in]		pointer to index first element.
 	len[in]			pointer to length of index,
-						which must be lower or equal
-						than the length of vectors
-						T1, SW->ptr, T2, E, MX and SX.
+					which must be lower or equal
+					than the length of vectors
+					T1, SW->ptr, T2, E, MX and SX.
 	H[in]			pointer to H first element.
 	kfunc[in]		pointer to kernel density function.
 	MX[out]			pointer to MX vector.
 	SX[out]			pointer to SX vector.
-	b[in]			pointer to row index.
+	t[in]			pointer to thread number.
+	WORK[out]		pointer to array of transLSW structures.
 
 Return value:
 	This function doesn't return a value.
@@ -304,7 +305,6 @@ Remarks:
 	Vector index must indicate the permutation of vector T2
 		sorted by ascending order.
 	Vectors T1, SW->ptr, T2, E, MX and SX must have the same length.
-	If (*b < 1) the code runs in parallel among the available threads.
 */
 
 static void LSmeasuresI(
@@ -318,15 +318,20 @@ static void LSmeasuresI(
 	Kfunc kfunc,
 	doubleCP MX,
 	doubleCP SX,
-	CintCP b)
+	CintCP t,
+	transLSW *WORK)
 {
 	#ifdef _OPENMP
-	#pragma omp parallel if(*b < 1) num_threads(global_num_threads)
+	#pragma omp parallel if( !omp_in_parallel() ) num_threads(global_num_threads)
 	#endif
 	{
 		register int i, j;
+		int tid = *t;
 		double aux;
-		double *K = (double*)malloc( *len*sizeof(double) ); // allocate memory block
+		#ifdef _OPENMP
+		if (omp_get_num_threads() != 1) tid = omp_get_thread_num(); // use the correct thread number
+		#endif
+		double *K = WORK[tid].K;
 		if (H[1] == H[0]) {
 			#ifdef _OPENMP
 			#pragma omp for
@@ -379,14 +384,13 @@ static void LSmeasuresI(
 				SX[index[i]] = sqrt(SX[index[i]]); // compute standard deviation
 			}
 		}
-		free(K); // free memory block
 	}
 	return;
 } // LSmeasuresI
 
 /*
 Author:
-	Artur Agostinho Araujo <artur.stat@gmail.com>
+	Artur Araujo <artur.stat@gmail.com>
 
 Description:
 	Computes the transition probabilities:
@@ -402,7 +406,7 @@ Parameters:
 	T2[in]			pointer to T2 first element.
 	E[in]			pointer to E first element.
 	index0[in]		pointer to index0 first element.
-	index1[inout]	pointer to index1 first element.
+	index1[inout]		pointer to index1 first element.
 	nt[in]			pointer to length of UT and number of rows of P.
 	UT[in]			pointer to unique times vector.
 	nb[in]			pointer to number of rows of P.
@@ -410,6 +414,8 @@ Parameters:
 	b[in]			pointer to row index.
 	kfunc[in]		pointer to kernel density function.
 	H[in]			pointer to H first element.
+	t[in]			pointer to thread number.
+	WORK[out]		pointer to array of transLSW structures.
 
 Return value:
 	This function doesn't return a value.
@@ -436,12 +442,14 @@ static void transLSI(
 	double P[*nb*(*nt)*4],
 	CintCP b,
 	Kfunc kfunc,
-	Cdouble H[2])
+	Cdouble H[2],
+	CintCP t,
+	transLSW *WORK)
 {
 	register int i;
 	int j, k, x, e0, e1;
 	double aux[2], p;
-	double *W = (double*)malloc( *len*sizeof(double) ); // allocate memory block
+	double *W = WORK[*t].a;
 	stype SW; // declare stype structure
 	SW.type = INT_PTR; // type is an int pointer
 	SW.ptr.integer = E1; // hold E1 pointer in ptr union
@@ -474,11 +482,11 @@ static void transLSI(
 		W[index0[j]] *= aux[0]; // compute and save weight
 		aux[0] *= aux[1]; // compute and save factor needed for next iteration
 	}
-	double *MX = (double*)malloc( *len*sizeof(double) ); // allocate memory block
-	double *SX = (double*)malloc( *len*sizeof(double) ); // allocate memory block
-	LSmeasuresI(T1, &SW, T2, E, index1, len, H, kfunc, MX, SX, b); // compute location-scale measures
-	double *EX = (double*)malloc( *len*sizeof(double) ); // allocate memory block
-	double *SV = (double*)malloc( *len*sizeof(double) ); // allocate memory block
+	double *MX = WORK[*t].b;
+	double *SX = WORK[*t].c;
+	LSmeasuresI(T1, &SW, T2, E, index1, len, H, kfunc, MX, SX, t, WORK); // compute location-scale measures
+	double *EX = WORK[*t].E0B;
+	double *SV = WORK[*t].E1B;
 	for (i = 0; i < *len; i++) EX[index1[i]] = (T2[index1[i]]-MX[index1[i]])/SX[index1[i]];
 	order_d(EX, index1, *len, FALSE, FALSE, SV); // get permuation
 	kmsurv(len, EX, E, index1, len, SV); // compute survival probabilities
@@ -509,11 +517,6 @@ static void transLSI(
 			x = e1;
 		}
 	}
-	free(MX); // free memory block
-	free(SX); // free memory block
-	free(EX); // free memory block
-	free(SV); // free memory block
-	free(W); // free memory block
 	#ifdef _OPENMP
 	#pragma omp parallel for if(*b < 1) num_threads(global_num_threads) private(i) ordered
 	#endif
@@ -538,7 +541,7 @@ static void transLSI(
 
 /*
 Author:
-	Artur Agostinho Araujo <artur.stat@gmail.com>
+	Artur Araujo <artur.stat@gmail.com>
 
 Description:
 	Computes a transition probability array
@@ -546,16 +549,16 @@ Description:
 
 Parameters:
 	object			an object of class 'LS'.
-	UT				unique times vector.
-	h				a vector of bandwidth values with length two.
-	nh				number of bandwidth values
-						to test by cross-validation.
-	ncv				number of cross-validation samples.
+	UT			unique times vector.
+	h			a vector of bandwidth values with length two.
+	nh			number of bandwidth values
+					to test by cross-validation.
+	ncv			number of cross-validation samples.
 	window			a string indicating the desired window or kernel.
 	nboot			number of bootstrap samples.
 	bootcv			if TRUE cross-validation is done for each bootstrap sample.
-	cvfull			if TRUE cross-validation is done for both location and scale
-						functions.
+	cvfull			if TRUE cross-validation is done for both location and
+					scale functions.
 
 Return value:
 	Returns a list where the first element is a
@@ -582,62 +585,89 @@ SEXP TransPROBLS(
 	E1 = VECTOR_ELT(data, 1);
 	S = VECTOR_ELT(data, 2);
 	E = VECTOR_ELT(data, 3);
-	int len = GET_LENGTH(T1), nt = GET_LENGTH(UT), b;
+	int len = GET_LENGTH(T1), nt = GET_LENGTH(UT), b, t, nth = 1;
 	double *T2 = (double*)malloc( len*sizeof(double) ); // allocate memory block
+	if (T2 == NULL) error("TransPROBLS: No more memory\n");
 	for (b = 0; b < len; b++) T2[b] = REAL(S)[b]-REAL(T1)[b];
 	Kfunc kfunc = kchar2ptr(window); // declare and get pointer to function
 	SEXP P, H, list;
 	PROTECT( P = alloc3DArray(REALSXP, *INTEGER(nboot), nt, 4) );
 	PROTECT( H = NEW_NUMERIC(2) );
 	PROTECT( list = NEW_LIST(2) );
-	int *index0 = (int*)malloc( len*sizeof(int) ); // allocate memory block
-	int *index1 = (int*)malloc( len*sizeof(int) ); // allocate memory block
-	double *WORK0 = (double*)malloc( len*sizeof(double) ); // allocate memory block
-	int *WORK1 = (int*)malloc( len*sizeof(int) ); // allocate memory block
+	transLSW *WORK = (transLSW*)malloc( global_num_threads*sizeof(transLSW) ); // allocate memory block
+	if (WORK == NULL) error("TransPROBLS: No more memory\n");
+	for (t = 0; t < global_num_threads; t++) { // allocate per thread memory
+		if ( ( WORK[t].sample0 = (int*)malloc( len*sizeof(int) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].sample1 = (int*)malloc( len*sizeof(int) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].unique0 = (int*)malloc( len*sizeof(int) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].K = (double*)malloc( len*sizeof(double) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].MX = (double*)malloc( len*sizeof(double) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].a = (double*)malloc( len*sizeof(double) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].b = (double*)malloc( len*sizeof(double) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].c = (double*)malloc( len*sizeof(double) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].E0B = (double*)malloc( len*sizeof(double) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( WORK[t].E1B = (double*)malloc( len*sizeof(double) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+	}
+	if (*INTEGER(nboot) > 1) nth = global_num_threads;
+	int **index0 = (int**)malloc( nth*sizeof(int*) ); // allocate memory block
+	if (index0 == NULL) error("TransPROBLS: No more memory\n");
+	int **index1 = (int**)malloc( nth*sizeof(int*) ); // allocate memory block
+	if (index1 == NULL) error("TransPROBLS: No more memory\n");
+	for (t = 0; t < nth; t++) { // allocate per thread memory
+		if ( ( index0[t] = (int*)malloc( len*sizeof(int) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+		if ( ( index1[t] = (int*)malloc( len*sizeof(int) ) ) == NULL ) error("TransPROBLS: No more memory\n");
+	}
 	double HC[2];
 	b = 0; // b = len, put it back to 0 or a crash might occur
-	indx_ii(index0, index1, &len); // initialize indexes
-	crossValid(REAL(T1), INTEGER(E1), REAL(S), INTEGER(E), T2, index0, &len, REAL(h), INTEGER(nh), INTEGER(ncv), LOGICAL(cvfull), kfunc, &b, HC); // compute bandwidth
-	order_di(REAL(T1), INTEGER(E1), index0, len, FALSE, FALSE, TRUE, WORK0, WORK1); // get permuation
-	order_d(T2, index1, len, FALSE, FALSE, WORK0); // get permuation
-	transLSI(&len, REAL(T1), INTEGER(E1), T2, INTEGER(E), index0, index1, &nt, REAL(UT), INTEGER(nboot), REAL(P), &b, kfunc, HC); // compute transition probabilities
-	free(index0); // free memory block
-	free(index1); // free memory block
-	free(WORK0); // free memory block
-	free(WORK1); // free memory block
+	t = 0; // t = nth, put it back to 0 or a crash might occur
+	indx_ii(&len, index0[0], index1[0]); // initialize indexes
+	crossValid(REAL(T1), INTEGER(E1), REAL(S), INTEGER(E), T2, index0[0], &len, REAL(h), INTEGER(nh), INTEGER(ncv), LOGICAL(cvfull), kfunc, HC, &t, WORK); // compute bandwidth
+	order_di(REAL(T1), INTEGER(E1), index0[0], len, FALSE, FALSE, TRUE, WORK[0].K, WORK[0].unique0); // get permuation
+	order_d(T2, index1[0], len, FALSE, FALSE, WORK[0].K); // get permuation
+	transLSI(&len, REAL(T1), INTEGER(E1), T2, INTEGER(E), index0[0], index1[0], &nt, REAL(UT), INTEGER(nboot), REAL(P), &b, kfunc, HC, &t, WORK); // compute transition probabilities
 	REAL(H)[0] = HC[0];
 	REAL(H)[1] = HC[1];
 	if (*INTEGER(nboot) > 1) {
 		#ifdef _OPENMP
-		#pragma omp parallel num_threads(global_num_threads) private(b) firstprivate(HC)
+		#pragma omp parallel num_threads(global_num_threads) private(b, t) firstprivate(HC)
 		#endif
 		{
-			int *index0 = (int*)malloc( len*sizeof(int) ); // allocate memory block
-			int *index1 = (int*)malloc( len*sizeof(int) ); // allocate memory block
-			double *WORK0 = (double*)malloc( len*sizeof(double) ); // allocate memory block
-			int *WORK1 = (int*)malloc( len*sizeof(int) ); // allocate memory block
 			#ifdef _OPENMP
-			unsigned int iseed = (unsigned int)time(NULL) ^ (unsigned int)omp_get_thread_num(); // save per thread seed
+			t = omp_get_thread_num();
 			#else
-			unsigned int iseed = (unsigned int)time(NULL);
+			t = 0;
 			#endif
-			srand(iseed); // set seed
 			#ifdef _OPENMP
 			#pragma omp for
 			#endif
 			for (b = 1; b < *INTEGER(nboot); b++) {
-				boot_ii(index0, index1, &len); // bootstrap indexes
-				if ( *LOGICAL(bootcv) ) crossValid(REAL(T1), INTEGER(E1), REAL(S), INTEGER(E), T2, index0, &len, REAL(h), INTEGER(nh), INTEGER(ncv), LOGICAL(cvfull), kfunc, &b, HC); // compute bandwidth
-				order_di(REAL(T1), INTEGER(E1), index0, len, FALSE, FALSE, TRUE, WORK0, WORK1); // get permuation
-				order_d(T2, index1, len, FALSE, FALSE, WORK0); // get permuation
-				transLSI(&len, REAL(T1), INTEGER(E1), T2, INTEGER(E), index0, index1, &nt, REAL(UT), INTEGER(nboot), REAL(P), &b, kfunc, HC); // compute transition probabilities
+				boot_ii(RngArray[t], &len, index0[t], index1[t]); // bootstrap indexes
+				if ( *LOGICAL(bootcv) ) crossValid(REAL(T1), INTEGER(E1), REAL(S), INTEGER(E), T2, index0[t], &len, REAL(h), INTEGER(nh), INTEGER(ncv), LOGICAL(cvfull), kfunc, HC, &t, WORK); // compute bandwidth
+				order_di(REAL(T1), INTEGER(E1), index0[t], len, FALSE, FALSE, TRUE, WORK[t].K, WORK[t].unique0); // get permuation
+				order_d(T2, index1[t], len, FALSE, FALSE, WORK[t].K); // get permuation
+				transLSI(&len, REAL(T1), INTEGER(E1), T2, INTEGER(E), index0[t], index1[t], &nt, REAL(UT), INTEGER(nboot), REAL(P), &b, kfunc, HC, &t, WORK); // compute transition probabilities
 			}
-			free(index0); // free memory block
-			free(index1); // free memory block
-			free(WORK0); // free memory block
-			free(WORK1); // free memory block
 		}
 	}
+	for (t = nth-1; t >= 0; t--) {
+		free(index0[t]); // free memory block
+		free(index1[t]); // free memory block
+	}
+	free(index0); // free memory block
+	free(index1); // free memory block
+	for (t = global_num_threads-1; t >= 0; t--) {
+		free(WORK[t].sample0); // free memory block
+		free(WORK[t].sample1); // free memory block
+		free(WORK[t].unique0); // free memory block
+		free(WORK[t].K); // free memory block
+		free(WORK[t].MX); // free memory block
+		free(WORK[t].a); // free memory block
+		free(WORK[t].b); // free memory block
+		free(WORK[t].c); // free memory block
+		free(WORK[t].E0B); // free memory block
+		free(WORK[t].E1B); // free memory block
+	}
+	free(WORK); // free memory block
 	free(T2); // free memory block
 	SET_ELEMENT(list, 0, P);
 	SET_ELEMENT(list, 1, H);
